""" Module doc """

# Version of ReduxPy package
__version__ = "0.1.2"

from typing import Tuple
from ._internal.store import create_store
from ._internal.action import create_action, select_action_payload, of_type
from ._internal.reducer import handle_actions, combine_reducers
from ._internal.feature import of_init_feature, create_feature_module, select_feature
from ._internal.epic import combine_epics
from ._internal.types import Action, Reducer, Epic, ReduxRootStore, ReduxFeatureModule, StateType
from ._internal.selectors import select

__all__: Tuple[str, ...] = (
    'Action',
    'combine_epics',
    'combine_reducers',
    'create_action',
    'create_feature_module',
    'create_store',
    'Epic',
    'handle_actions',
    'of_init_feature',
    'of_type',
    'Reducer',
    'ReduxFeatureModule',
    'ReduxRootStore',
    'select_action_payload',
    'select_feature',
    'select',
    'StateType',
)



"""
    Implements reducer related function
"""

from typing import Iterable, Mapping, MutableMapping, Optional, Tuple

from .action import select_action_type
from .types import Action, Reducer, StateType


def default_reducer(initial_state: Optional[StateType]) -> Reducer:
    """ Creates a reducer that returns the original state or the default state.

        Args:
            inital_state: optional initial state used as a fallback

        Returns:
            A reducer that reduces the current state or the initial state as a fallback

    """

    def _reducer(state: StateType, _: Action) -> Optional[StateType]:
        """ Reducer that returns the state or the initial state

            Args:
                state: the current state
                action: the action, will be ignored by the function

            Returns:
                The state of the initial state as a fallback

        """
        return state if state else initial_state

    return _reducer


def handle_actions(
    action_map: Mapping[str, Reducer], initial_state: Optional[StateType] = None,
) -> Reducer:
    """ Creates a new reducer from a mapping of action name to reducer for that action.

        Args:
            action_map: mapping from action name to reducer for that action
            initial_state: optional initial state used if no reducer matches

        Returns:
            A reducer function that handles the actions in the map

    """
    def_reducer = default_reducer(initial_state)

    def _reducer(state: StateType, action: Action) -> Optional[StateType]:
        """ Applies the mapped reducer or returns the default state

            Args:
                state: current state
                action: action to apply against the current state

            Returns:
                the updated state
        """
        return action_map.get(select_action_type(action),
                              def_reducer)(state, action)

    return _reducer


def combine_reducers(
        reducers: Mapping[str, Reducer]) -> Reducer[Mapping[str, StateType]]:
    """ Creates a new reducer from a mapping of reducers.

        Args:
            reducers: the mapping from state partition to reducer

        Returns:
            A reducer that dispatches actions against each of the mapped reducers

    """
    items: Iterable[Tuple[str, Reducer]] = tuple(reducers.items())

    def _combine(state: Mapping[str, StateType],
                 action: Action) -> Mapping[str, StateType]:
        """ Updates the state object from the reducer mappings. """
        result = state if state else {}
        mutable: Optional[MutableMapping[str, StateType]] = None
        for key, value in items:
            current = result.get(key)
            updated = value(current, action)
            if not current is updated:
                if not mutable:
                    mutable = dict(result)
                    result = mutable
                mutable[key] = updated
        return result

    return _combine



"""
    Type definitions
"""
from typing import (Any, Callable, Iterable, Mapping, NamedTuple, Optional,
                    TypeVar)

from rx.core import Observable

StateType = TypeVar("StateType")

PayloadType = TypeVar("PayloadType")

ReduxRootState = Mapping[str, StateType]


class Action(NamedTuple):
    """ Action implementation that takes a payload """
    type: str
    payload: Any


Epic = Callable[[Observable, Observable],
                Observable]

Reducer = Callable[[StateType, Action], StateType]


class ReduxFeatureModule(NamedTuple):
    """ Defines the feature module. The ID identifies the section in the state and
        is also used to globally discriminate features.

        After instantiating a feature store the store will fire an initialization action
        for that feature. Use `ofInitFeature` to register for these initialization actions.
    """
    id: str
    reducer: Optional[Reducer]
    epic: Optional[Epic]
    dependencies: Iterable[Any]


class ReduxRootStore(NamedTuple):
    """ Implementation of a store that manages sub-state as features. Features are added
        to the store automatically, when required by the select method.
    """
    as_observable: Callable[[], Observable]
    dispatch: Callable[[Action], None]
    add_feature_module: Callable[[ReduxFeatureModule], None]
    on_next: Callable[[Action], None]
    on_completed: Callable[[], None]
